open Common;;
open Types;;
open Set;;
open Syntax;;
open Tprinter;;
open Typing;;
open Value;;
open Primitive;;
open Lexbuf;;

type frame =
   FrApplL of env * expr
 | FrApplR of value
 | FrLet of ident list * env * value list * expr list * expr
 | FrLetRec of recval list * env * value list * expr list * expr
 | FrIf of env * expr * expr
 | FrConsL of label * env * expr
 | FrConsR of value * label
 | FrRes of label
 | FrSel of label;;

let ma_compute =
 let stack = Array.make 20000 (FrRes "")
 and sp = ref (-1)
 and env = ref [] in

 let rec code = function
    Appl (m, n) -> incr sp;
                   stack.(!sp) <- FrApplL (!env, n);
                   code m
  | Let (binds, n) -> let (xs, ms) = List.split binds in
                      incr sp;
                      stack.(!sp) <- FrLet (xs, !env, [], List.tl ms, n);
                      code (List.hd ms)
  | LetRec (binds, n) ->
             let (xs, ms) = List.split binds in
             let rs = List.map (fun _ -> ref None) xs in
             let e' = (List.map2 (fun x r -> (x, RecVal r)) xs rs) @ !env in
             incr sp;
             stack.(!sp) <- FrLetRec (rs, e', [], List.tl ms, n);
             env := e';
             code (List.hd ms)
  | Fun (x, _, m) -> value (Closure (x, m, !env))
  | Ident x -> value (search_env x !env)
  | IntE i -> value (Int i)
  | StringE s -> value (String s)
  | If (b, m, n) -> incr sp;
                    stack.(!sp) <- FrIf (!env, m, n);
                    code b
  | REmpty -> value (Record [])
  | RCons (m, l, n) -> incr sp;
                       stack.(!sp) <- FrConsL (l, !env, n);
                       code m
  | RSel (m, l) -> incr sp;
                   stack.(!sp) <- FrSel l;
                   code m
  | RRes (m, l) -> incr sp;
                   stack.(!sp) <- FrRes l;
                   code m
  | TypeConstr (m, _) -> code m

 and value = function
    RecVal r ->
     begin match !r with
         Some v -> value v
       | None -> begin match stack.(!sp) with
                    FrApplR (Closure (x, m, e)) -> decr sp;
                                                   env := (x, RecVal r)::e;
                                                   code m
                  | FrApplR (Primitive f) -> decr sp;
                                             value (f (RecVal r))
                  | _ -> failwith "RecVal-error"
                 end
     end
  | v -> if !sp < 0 then v else
         match stack.(!sp) with
            FrApplL (e, n) -> stack.(!sp) <- FrApplR v;
                              env := e;
                              code n
          | FrApplR (Closure (x, m, e)) -> decr sp;
                                           env := (x, v)::e;
                                           code m
          | FrApplR (Primitive f) -> decr sp;
                                     value (f v)
          | FrLet (xs, e, vs, m::ms, n) ->
                 stack.(!sp) <- FrLet (xs, e, v::vs, ms, n);
                 env := e;
                 code m
          | FrLet (xs, e, vs, [], n) ->
                 decr sp;
                 env := List.combine xs (List.rev (v::vs)) @ e;
                 code n
          | FrLetRec (rs, e, vs, m::ms, n) ->
                 stack.(!sp) <- FrLetRec (rs, e, v::vs, ms, n);
                 env := e;
                 code m
          | FrLetRec (rs, e, vs, [], n) ->
                 List.iter2 (fun r v -> r := Some v) rs (List.rev (v::vs));
                 decr sp;
                 env := e;
                 code n
          | FrIf (e, m, n) -> begin match v with
                                 Bool b -> decr sp;
                                           env := e;
                                           code (if b then m else n)
                               | _ -> failwith "If-error"
                              end
          | FrConsL (l, e, n) -> stack.(!sp) <- FrConsR (v, l);
                                 env := e;
                                 code n
          | FrConsR (Record r, l) -> decr sp;
                                     value (Record (ins_record l v r))
          | FrSel l -> begin match v with
                          Record r -> decr sp;
                                      value (sel_record l r)
                        | _ -> failwith "RSel-error"
                       end
          | FrRes l -> begin match v with
                          Record r -> decr sp;
                                      value (Record (supp_record l r))
                        | _ -> failwith "RRes-error"
                       end
          | FrApplR _ -> failwith "Lambda-error"
          | FrConsR _ -> failwith "RCons-error"

 in
 fun e m -> sp := -1;
            env := e;
            try
             code m
            with Invalid_argument "Array.set" -> runtime_msg "Stack overflow";;

let env = ref start_env;;
let envT = ref start_envT;;

let rec eval_instr = function
   Def binds -> if !enable_typing then
                begin
                 let (xs, ms) = List.split binds in
                 let ts = List.map (fun m -> fst (infer !envT m)) ms in
		 let vs = List.map (ma_compute !env) ms in
		 env := (List.combine xs vs) @ !env;
		 let ts' = List.map2 (fun t m -> generalize t !envT m) ts ms in
		 envT := (List.combine xs ts') @ !envT;
		 iter3 (fun x t' v -> print_string "val "; print_string x;
			              print_string " : "; print_scheme t';
				      print_string " = "; print_value v;
				      print_newline (); start_printing ())
		       xs ts' vs
                end
                else
                begin
                 let (xs, ms) = List.split binds in
		 let vs = List.map (ma_compute !env) ms in
		 env := (List.combine xs vs) @ !env;
		 List.iter2 (fun x v -> print_string "val "; print_string x;
			                print_string " : unsafe = ";
                                        print_value v; print_newline ();
                                        start_printing ())
		            xs vs
                end
 | DefRec binds -> if !enable_typing then
                   begin
                    let (xs, ms) = List.split binds in
                    let bs = List.map (fun _ -> tvar_cell ()) xs in
                    let envT' = (List.map2 (fun x b -> (x, (poly_empty, b)))
		 	 	  	    xs bs) @ !envT in
                    let (ts, fs) = List.split (List.map (infer envT') ms) in
		    (* les deux mgu ci-dessous peuvent renvoyer un
		       Type_runtime qui devrait tre converti en Runtime
		       normalement en affichant le lieu de l'erreur... *)
                    List.iter2 mgu_typ ts bs;
                    List.iter (fun x ->
		      List.iter (fun f -> mgu_degree_one (f x)) fs) xs;
                    let rs = List.map (fun _ -> ref None) xs in
                    env := (List.map2 (fun x r -> (x, RecVal r)) xs rs) @ !env;
                    let vs = List.map (ma_compute !env) ms in
                    List.iter2 (fun r v -> r := Some v) rs vs;
                    let ts' = List.map2 (fun t m -> generalize t !envT m)
                                        ts ms in
                    envT := (List.combine xs ts') @ !envT;
		    iter3 (fun x t' v -> print_string "val "; print_string x;
		 	                 print_string " : "; print_scheme t';
				         print_string " = "; print_value v;
				         print_newline (); start_printing ())
		          xs ts' vs
                   end
                   else
                   begin
                    let (xs, ms) = List.split binds in
                    let rs = List.map (fun _ -> ref None) xs in
                    env := (List.map2 (fun x r -> (x, RecVal r)) xs rs) @ !env;
                    let vs = List.map (ma_compute !env) ms in
                    List.iter2 (fun r v -> r := Some v) rs vs;
		    List.iter2 (fun x v -> print_string "val "; print_string x;
		 	                   print_string " : unsafe = ";
				           print_value v; print_newline ();
                                           start_printing ())
		               xs vs
                   end
 | Expr m -> if !enable_typing then
             begin
              let t = fst (infer !envT m) in
              let v = ma_compute !env m in
              let t' = generalize t !envT m in
              print_string "- : "; print_scheme t';
              print_string " = "; print_value v; print_newline ()
             end
             else eval_instr (Unsafe m)
 | Type m -> let t = fst (infer !envT m) in
             let t' = generalize t !envT m in
             print_string "- : "; print_scheme t'; print_string " = <type>";
             print_newline ()
 | Unsafe m -> let v = ma_compute !env m in
               print_string "- : unsafe = "; print_value v; print_newline ()
 | Use s -> interpret_file s

and interpret_file s =
 try
  let ch = open_in s in
  interpret_channel false ch;
  close_in ch
 with
    Sys_error _ -> print_string "Error while reading "; print_string s;
                   print_newline ();
  | Lexer.Eof -> ()

and interpret_channel interactive ch =
 try
   let lexbuf = Lexing.from_channel ch in
   while true do
     try
       start_printing ();
       if interactive then (print_string "> "; flush stdout);
       let instr = try Parser.main Lexer.token lexbuf
                   with Parsing.Parse_error -> runtime_msg "Syntax error" in
       if interactive then print_newline ();
       eval_instr instr;
       if interactive then print_newline ()
     with Failure s -> print_string "Internal error: "; print_string s;
                       print_newline ();
                       if interactive then empty_lexbuf lexbuf
                                      else raise Lexer.Eof
          (* Type_runtime ne peut provenir que du let rec au toplevel,
	     voir ci-dessus *)
        | Runtime f | Type_runtime f ->
             f (); if interactive then (print_newline (); empty_lexbuf lexbuf)
                                  else raise Lexer.Eof
   done
 with
   Quit -> exit 0;;
