open Common;;
open Types;;
open Set;;

type expr =
   Appl of expr * expr
 | Let of (ident * expr) list * expr
 | LetRec of (ident * expr) list * expr
 | Fun of ident * typ cell option * expr
 | Ident of ident
 | IntE of int
 | StringE of string
 | If of expr * expr * expr
 | REmpty
 | RCons of expr * label * expr
 | RSel of expr * label
 | RRes of expr * label
 | TypeConstr of expr * typ cell;;

type instr =
   Def of (ident * expr) list
 | DefRec of (ident * expr) list
 | Expr of expr
 | Type of expr
 | Unsafe of expr
 | Use of string;;

let fv_expr =
 let rec loop l = function
    Appl (m, n) -> loop (loop l m) n
  | Let (binds, n) -> let (xs,ms) = List.split binds in
                      List.fold_left loop (set_minus (loop l n) xs) ms
  | LetRec (binds, n) -> let (xs,ms) = List.split binds in
                         set_minus (loop (List.fold_left loop l ms) n) xs
  | Fun (x, _, m) -> set_minus (loop l m) [x]
  | Ident x -> x::l
  | IntE _ -> l
  | StringE _ -> l
  | If (b, m, n) -> loop (loop (loop l b) m) n
  | REmpty -> l
  | RCons (m, _, n) -> loop (loop l m) n
  | RSel (m, _) -> loop l m
  | RRes (m, _) -> loop l m
  | TypeConstr (m, _) -> loop l m
 in loop [];;

let rec is_pure = function
   Appl _ -> false
 | Let(binds,n) -> List.for_all is_pure (List.map snd binds) && is_pure n
 | LetRec (binds,n) -> List.for_all is_pure (List.map snd binds) && is_pure n
 | Ident _ -> true
 | Fun _ -> true
 | IntE _ -> true
 | StringE _ -> true
 | If (b,m,n) -> is_pure b && is_pure m && is_pure n
 | REmpty -> true
 | RCons (m,_,n) -> is_pure m && is_pure n
 | RSel (m,_) -> is_pure m
 | RRes (m,_) -> is_pure m
 | TypeConstr (m, _) -> is_pure m;;

let iter3 f l1 l2 l3 =
 let rec loop = function
    (x1::r1, x2::r2, x3::r3) -> f x1 x2 x3; loop (r1, r2, r3)
  | ([], [], []) -> ()
  | _ -> failwith "Iter3-error"
 in loop (l1, l2, l3);;

let rec print_expr par = function
   Appl (m,n) -> if par then print_char '(';
                 print_expr true m; print_string " "; print_expr true n;
                 if par then print_char ')'
 | Let (l,m) -> if par then print_char '(';
                print_string "let "; print_expr_list l; print_string " in ";
                print_expr false m;
                if par then print_char ')'
 | LetRec (l,m) -> if par then print_char '(';
                   print_string "let rec "; print_expr_list l;
                   print_string " in "; print_expr false m;
                   if par then print_char ')'
 | Fun (x,_,m) -> if par then print_char '(';
                  print_string "fun "; print_string x; print_string " -> ";
                  print_expr false m;
                  if par then print_char ')'
 | Ident x -> print_string x
 | IntE n -> print_int n
 | StringE s -> print_string "\""; print_string s; print_string "\""
 | If (b,m,n) -> if par then print_char '(';
                 print_string "if "; print_expr false b; print_string " then ";
                 print_expr false m; print_string " else "; print_expr false n;
                 if par then print_char ')'
 | REmpty -> print_string "{ }"
 | RCons (REmpty,l,n) -> print_string "{ "; print_string l;
                         print_string " = "; print_expr false n;
                         print_string " }"
 | RCons (m,l,n) -> print_string "{ "; print_expr false m; print_string ", ";
                    print_string l; print_string " = "; print_expr false n;
                    print_string " }"
 | RSel (m,l) -> print_expr true m; print_string "."; print_string l
 | RRes (m,l) -> print_expr true m; print_string "\\"; print_string l
 | TypeConstr (m,_) -> print_expr par m

and print_expr_list = function
   [] -> failwith "Empty let-binding-list"
 | [(x,m)] -> print_string x; print_string " = "; print_expr false m
 | ((x,m)::l) -> print_string x; print_string " = "; print_expr false m;
                 print_string " and "; print_expr_list l
;;
