open Common;;

type constr = label list;;

type typ =
   IntT
 | BoolT
 | UnitT
 | StringT
 | TVar
 | FunT of typ cell * degree cell * typ cell
 | RecordT of record cell
 | ConstrT of string * typ cell

and record =
   EmptyT
 | LabelT of label * typ cell
 | NodeT of record cell * record cell
 | RVar of constr

and degree =
   Zero
 | One
 | DVar of degree cell list ref * degree cell list ref
 | Inf of degree cell * degree cell

and 'a cell = { data : 'a; mutable repr : 'a cell };;

let new_cell d =
 let rec u = { data = d; repr = u } in u;;

let rec repr u =
 if u.repr == u
 then u
 else let v = repr u.repr in u.repr <- v; v;;

let int_cell = new_cell IntT;;
let bool_cell = new_cell BoolT;;
let unit_cell = new_cell UnitT;;
let string_cell = new_cell StringT;;
let tvar_cell () = new_cell TVar;;
let fun_cell (u1, u2, u3) = new_cell (FunT (u1, u2, u3));;
let record_cell u = new_cell (RecordT u);;
let constr_cell (s, u) = new_cell (ConstrT (s, u));;

let empty_cell = new_cell EmptyT;;
let label_cell (l, u) = new_cell (LabelT (l, u));;
let node_cell (u1, u2) = new_cell (NodeT (u1, u2));;
let rvar_cell constr = new_cell (RVar constr);;

let zero_cell = new_cell Zero;;
let one_cell = new_cell One;;
let dvar_cell () = new_cell (DVar (ref [], ref []));;
let inf_cell (u1, u2) = new_cell (Inf (u1, u2));;

let ins_rec (l,t) =
 let rec loop = function
    [] -> [(l,t)]
  | ((l',t')::r) -> if l < l' then (l,t)::(l',t')::r
                    else if l = l' then failwith "Ins_rec-Error"
                    else (l',t')::(loop r)
 in loop;;

let rec_fields u =
 let var = ref None and l_lab = ref [] in
 let rec loop u =
  let u = repr u in
  match u.data with
    EmptyT -> ()
  | LabelT (l,t) -> l_lab := ins_rec (l,t) !l_lab
  | NodeT (u1,u2) -> loop u1; loop u2
  | RVar _ -> begin match !var with None -> var := Some u
                                  | _ -> failwith "Rec_fields-Error"
              end
 in loop u; (!var,!l_lab);;

let make_rec base =
 let rec loop = function
    [] -> base
  | ((l,t)::r) -> node_cell (label_cell (l,t), loop r)
 in loop;;

let lab_not_allowed u =
 match u.data with
    RVar l -> l
  | _ -> failwith "Lab_not_allowed-Error";;

let infs_sups u =
 match u.data with
    DVar (l1,l2) -> (l1,l2)
  | _ -> failwith "Infs_sups-Error";;

type poly_var = { tvar : typ cell list;
                  rvar : record cell list;
                  dvar : degree cell list };;

type scheme = poly_var * typ cell;;

type envT = (ident * scheme) list;;

let search_env x (e : envT) =
 try
  List.assoc x e
 with Not_found -> runtime_msg ("Identifier " ^ x ^ " not defined");;

let rec mix_rec (l : (label * typ cell) list) l' = match (l,l') with
   ([],l) -> l
 | (l,[]) -> l
 | (x1::l1, x2::l2) -> if fst x1 < fst x2 then x1::(mix_rec l1 l')
                                          else x2::(mix_rec l l2);;
